# -* coding: utf-8 -*-

#################################################################################
# This file is part of EVALLIES.                                                #
#                                                                               #
# EVALLIES is a python package for lifelong learning speaker diarization.       #
# Home page: https://git-lium.univ-lemans.fr/Larcher/evallies                   #
#                                                                               #
# EVALLIES is free software: you can redistribute it and/or modify              #
# it under the terms of the GNU LLesser General Public License as               #
# published by the Free Software Foundation, either version 3 of the License,   #
# or (at your option) any later version.                                        #
#                                                                               #
# EVALLIES is distributed in the hope that it will be useful,                   #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 #
# GNU Lesser General Public License for more details.                           #
#                                                                               #
# You should have received a copy of the GNU Lesser General Public License      #
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.              #
#                                                                               #
#################################################################################
"""
Copyright 2020-2021 Anthony Larcher

    :mod:`lium_baseline.__init__`

"""

from .idmap import IdMap
from .ndx import Ndx
from .plotwindow import PlotWindow
from .key import Key
from .scores import Scores
from .detplot import DetPlot
from .detplot import effective_prior
from .detplot import logit_effective_prior
from .detplot import fast_minDCF


__author__ = "Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'
__credits__ = ["Anthony Larcher"]




# -* coding: utf-8 -*-

#################################################################################
# This file is part of EVALLIES.                                                #
#                                                                               #
# EVALLIES is a python package for lifelong learning speaker diarization.       #
# Home page: https://git-lium.univ-lemans.fr/Larcher/evallies                   #
#                                                                               #
# EVALLIES is free software: you can redistribute it and/or modify              #
# it under the terms of the GNU LLesser General Public License as               #
# published by the Free Software Foundation, either version 3 of the License,   #
# or (at your option) any later version.                                        #
#                                                                               #
# EVALLIES is distributed in the hope that it will be useful,                   #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 #
# GNU Lesser General Public License for more details.                           #
#                                                                               #
# You should have received a copy of the GNU Lesser General Public License      #
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.              #
#                                                                               #
#################################################################################
"""
Copyright 2020-2021 Anthony Larcher, Meysam Shamsi & Yevhenii Propkopalo

    :mod:`lium_baseline.interactive`

"""

import copy
import numpy
import pandas
import scipy
import os

from s4d.clustering.hac_utils import scores2distance
from scipy.cluster import hierarchy as hac
from scipy.spatial.distance import squareform

from .utils import s4d_to_allies
from ..user_simulation import MessageToUser
from ..user_simulation import Request
from s4d.clustering.hac_iv import information
from ..der_single import *

from s4d.diar import Diar
from s4d.scoring import DER
import json

def allies_write_diar(current_diar, filename):
    """

    """
    cdiar = copy.deepcopy(current_diar)
    for idx, seg in enumerate(cdiar):
        cdiar.segments[idx]['start'] = float(seg['start']) / 100.
        cdiar.segments[idx]['stop'] = float(seg['stop']) / 100.


    cdiar.sort(['show', 'start'])
    with open(filename, 'w', encoding="utf8") as fic:
        for line in Diar.to_string_seg(cdiar, time_float=True):
            fic.write(line)


def get_node_spkeakers(node_spk,number_cluster,link):
    """
    get a node and return the all sub cluster id

    :param node_spk: row of the linkage matrix
    :param number_cluster: number in link matrix
    :param link: linkage matrix
    :return: cluster_list: a list of clusters id in link
    """

    if node_spk >= number_cluster:
        cluster_list=get_node_spkeakers(link[int(node_spk-number_cluster)][0],number_cluster,link)+get_node_spkeakers(link[int(node_spk-number_cluster)][1],number_cluster,link)
    else:
        cluster_list=[int(node_spk)]
    return cluster_list


def ask_question(node_to_check, scores,scores_init, current_diar,current_vec, user, file_info,link,selection_criteria="longest"):
    """


    :param node_to_check: row of the linkage matrix to check with the user
    :param scores: the corresponding score object
    :param current_diar: the current diarization
    :param user: the user simulation
    :param link: linkage matrix
    :return: is_same_speaker: a boolean, True if the clustering at this node is approved by the user
    """
 
    # get one segment for each side of the node (if it's a leaf, it's easy, if qe already have a cluster we need to
    # list all segments in this cluster and then find the most central segment in the cluster from the score matrix
    # HERE WE NEED TO FIND TWO indices in the scores.scoremat matrix
    number_cluster=len(scores.modelset)
    node_list1=get_node_spkeakers(node_to_check[0],number_cluster,link)
    node_list2=get_node_spkeakers(node_to_check[1],number_cluster,link)
    #convert node_list to spk_list
    #import pdb
    #pdb.set_trace()     

    spk_list1=[]
    for n in node_list1:
        spk_list1.append(scores.modelset[n])
    spk_list2=[]
    for n in node_list2:
        spk_list2.append(scores.modelset[n])
    
    # From the indices, get the borders of the segments in the Diar object
    seg1=[0,0]
    seg2=[0,0]
    
    if selection_criteria == "longest":
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list1:
                if current_diar[i][4]-current_diar[i][3]>seg1[1]-seg1[0]:
                    seg1[1]=current_diar[i][4]
                    seg1[0]=current_diar[i][3]
            if current_diar[i][1] in spk_list2:
                if current_diar[i][4]-current_diar[i][3]>seg2[1]-seg2[0]:
                    seg2[1]=current_diar[i][4]
                    seg2[0]=current_diar[i][3]
    
    elif selection_criteria == "cluster_center":
        segs_cluster1={}
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list1:
                segs_cluster1[i]=current_vec.stat1[i]
        cluster1_center=numpy.mean(list(segs_cluster1.values()), axis=0)
        min_dis=float('inf')
        for seg in segs_cluster1:
            dis=scipy.spatial.distance.euclidean(segs_cluster1[seg],cluster1_center)
            if dis<min_dis:
                min_dis=dis
                seg1=[current_diar[int(seg)][3],current_diar[int(seg)][4]]
        segs_cluster2={}
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list2:
                segs_cluster2[i]=current_vec.stat1[i]
        cluster2_center=numpy.mean(list(segs_cluster2.values()), axis=0)
        min_dis=float('inf')
        for seg in segs_cluster2:
            dis=scipy.spatial.distance.euclidean(segs_cluster2[seg],cluster2_center)
            if dis<min_dis:
                min_dis=dis
                seg2=[current_diar[int(seg)][3],current_diar[int(seg)][4]]
                
    elif selection_criteria == "cluster_center_plda":
        
        segs_cluster1=[]
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list1:
                segs_cluster1.append(i)
        cluster1_similaritySum=[]
        for i in range(len(segs_cluster1)):
            cluster1_similaritySum.append(0)
            for j in range(len(scores_init.scoremat[i])):
                if j in segs_cluster1 and j!=i:
                     cluster1_similaritySum[i]+=scores_init.scoremat[i][j]
                        
        center1_seg=segs_cluster1[cluster1_similaritySum.index(max(cluster1_similaritySum))]
        seg1=[current_diar[int(center1_seg)][3],current_diar[int(center1_seg)][4]]        
        
        segs_cluster2=[]
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list2:
                segs_cluster2.append(i)
        cluster2_similaritySum=[]
        for i in range(len(segs_cluster2)):
            cluster2_similaritySum.append(0)
            for j in range(len(scores_init.scoremat[i])):
                if j in segs_cluster2 and j!=i:
                     cluster2_similaritySum[i]+=scores_init.scoremat[i][j]
                        
        center2_seg=segs_cluster2[cluster2_similaritySum.index(max(cluster2_similaritySum))]
        seg2=[current_diar[int(center2_seg)][3],current_diar[int(center2_seg)][4]]
        
    elif selection_criteria == "random":
        import random
        segs_cluster1=[]
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list1:
                segs_cluster1.append(i)
        random_seg=random.choice(segs_cluster1)
        seg1=[current_diar[int(random_seg)][3],current_diar[int(random_seg)][4]]
        segs_cluster2=[]
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list2:
                segs_cluster2.append(i)
        random_seg=random.choice(segs_cluster2)
        seg2=[current_diar[int(random_seg)][3],current_diar[int(random_seg)][4]]


        
    elif selection_criteria == "max":
        
        segs_cluster1=[]
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list1:
                segs_cluster1.append(i)
        
        segs_cluster2=[]
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list2:
                segs_cluster2.append(i)
                  
        min_similardis=float('inf')
        for i in segs_cluster1:
            for j in segs_cluster2:
                if scores_init.scoremat[i][j]< min_similardis:
                    min_similardis=scores_init.scoremat[i][j]
                    seg1=[current_diar[i][3],current_diar[i][4]]
                    seg2=[current_diar[j][3],current_diar[j][4]]
                             
        
    elif selection_criteria == "min":
        segs_cluster1=[]
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list1:
                segs_cluster1.append(i)
        
        segs_cluster2=[]
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list2:
                segs_cluster2.append(i)
                  
        max_similardis=-float('inf')
        for i in segs_cluster1:
            for j in segs_cluster2:
                if scores_init.scoremat[i][j]> max_similardis:
                    max_similardis=scores_init.scoremat[i][j]
                    seg1=[current_diar[i][3],current_diar[i][4]]
                    seg2=[current_diar[j][3],current_diar[j][4]]

    elif selection_criteria == "max_noBICHAC":

        segs_cluster1=node_list1
        segs_cluster2=node_list2
        
        #import pdb
        #pdb.set_trace() 
        max_similardis=-float('inf')
        for i in segs_cluster1:
            for j in segs_cluster2:
                if scores_init.scoremat[i][j]> max_similardis:
                    max_similardis=scores_init.scoremat[i][j]
                    seg1=[current_diar[i][3],current_diar[i][4]]
                    seg2=[current_diar[j][3],current_diar[j][4]]


    elif selection_criteria == "ideal":
        segs_cluster1=[]
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list1:
                segs_cluster1.append(i)
        segs_cluster2=[]
        for i in range(len(current_diar)):
            if current_diar[i][1] in spk_list2:
                segs_cluster2.append(i)
        #import pdb
        #pdb.set_trace()  
        root  ='/lium/buster1/prokopalo/EXPE/icassp_hal_baseline/data/ALLIES/eval'
        basename = current_diar[0]['show']
        ref = Diar.read_mdtm(root+'/mdtm/'+basename+'.mdtm')          
        
        if len(ref)!=len(current_diar):
            print ("Error ! number of segment in diar are not same ..."+current_diar[0]['show'])
        dict1 = {}      
        for indx in segs_cluster1:
            if len(ref)>indx:
                if ref[indx]['cluster'] in dict1.keys():
                    dict1[ref[indx]['cluster']]+=(ref[indx]['stop']-ref[indx]['start'])
                else:
                    dict1[ref[indx]['cluster']]=(ref[indx]['stop']-ref[indx]['start'])

        dict2 = {}
        for indx in segs_cluster2:
            if len(ref)>indx:
                if ref[indx]['cluster'] in dict2.keys():
                    dict2[ref[indx]['cluster']]+=(ref[indx]['stop']-ref[indx]['start'])
                else:
                    dict2[ref[indx]['cluster']]=(ref[indx]['stop']-ref[indx]['start'])

        if max(dict2, key = dict2.get) == max(dict1, key = dict1.get):
            return True , 0
        else:
            return False , 0

        #import pdb
        #pdb.set_trace()    

    # For each segments, get the center of the segment in seconds
    t1 =numpy.mean(seg1)/100
    t2 =numpy.mean(seg2)/100
    
    
    # Ask the question to the user
    message_to_user = MessageToUser(file_info,
                                    s4d_to_allies(current_diar),
                                    Request('same', t1, t2))
    hal, answer = user.validate(message_to_user)
    return answer.answer,hal


def check_std_change(node_to_check, scores, current_diar,current_vec,link,question,active=False):
    
    if not active:
        return True
    
    number_cluster=len(scores.modelset)
    node_list1=get_node_spkeakers(node_to_check[0],number_cluster,link)
    node_list2=get_node_spkeakers(node_to_check[1],number_cluster,link)

    spk_list1=[]
    for n in node_list1:
        spk_list1.append(scores.modelset[n])
    spk_list2=[]
    for n in node_list2:
        spk_list2.append(scores.modelset[n])
        
    segs_cluster_mrg={}
    for i in range(len(current_diar)):
        if current_diar[i][1] in spk_list1 or current_diar[i][1] in spk_list2:
            segs_cluster_mrg[i]=current_vec.stat1[i]
    mrg_situation=numpy.mean(numpy.std(list(segs_cluster_mrg.values()), axis=0))

    segs_cluster1_sep={}
    for i in range(len(current_diar)):
        if current_diar[i][1] in spk_list1:
            segs_cluster1_sep[i]=current_vec.stat1[i]
    std_sep1=numpy.std(list(segs_cluster1_sep.values()), axis=0)
    segs_cluster2_sep={}
    for i in range(len(current_diar)):
        if current_diar[i][1] in spk_list2:
            segs_cluster2_sep[i]=current_vec.stat1[i]
    std_sep2=numpy.std(list(segs_cluster2_sep.values()), axis=0)
    sep_situation=numpy.mean(numpy.mean([std_sep1,std_sep2], axis=0))
    
    if question =="separate":
        question_quality= sep_situation<mrg_situation
    elif question =="merge":
        question_quality= sep_situation>mrg_situation
    else:
        print("Question type is not valid ...")
        question_quality=True
        
    return question_quality


def check_der(current_diar,cluster_list,link,root='/lium/buster1/mshamsi/expe/icassp_hal_baseline/data/ALLIES/eval'):
    
    
    cluster_dict = dict()
    merge = list()
    i = 0

    while i < len(link):
#         print("change ... ",link[i])
        # the cluster_list of the 2 clusters
        if len(cluster_list)>int(link[i][0]) and len(cluster_list)>int(link[i][1]): 
            c0 = cluster_list[int(link[i][0])]
            c1 = cluster_list[int(link[i][1])]
            information(merge, i, c0, c1, link[i][2])
            if c1 in cluster_dict:
                # c0 is put in c1, and c1 is not empty
                cluster_dict[c1].append(c0)
            else:
                cluster_dict[c1] = [c0]
            if c0 in cluster_dict:
                # remove c0 key
                cluster_dict[c1] += cluster_dict[c0]
                cluster_dict.pop(c0)
            # add the speaker of the new cluster
            cluster_list.append(c1)
            current_diar.rename('cluster', [c0], c1)
        i += 1
    
    basename = current_diar[0]['show']
    ref = load_mdtm(root+'/mdtm/'+basename+'.mdtm')
    uem = load_uem (root+'/uem/'+basename+'.uem')
    
    if not os.path.isdir(f"./tmp/diar_tmp/"):
        os.makedirs(f"./tmp/diar_tmp/")
    
    allies_write_diar(current_diar, f"./tmp/diar_tmp/hyp_{basename}.mdtm")
    
    hyp = load_mdtm(f"./tmp/diar_tmp/hyp_{basename}.mdtm")
    
    (der, time, map)= compute_der(ref, hyp, uem, {}, 0.250)
    
    return der, time


def active_learning_tree(current_diar,
                         initial_diar,
                         current_vec,
                         scores,
                         scores_init,
                         threshold,
                         user,
                         file_info,
                         clustering_method="complete",
                         selection_method="longest",
                         no_more_clustering = False,
                         no_more_separation = False,
                         process_all_nodes = True,
                         conditional_questioning=False):
    """

    :param current_diar: the segmentation that comes out from the automatic pass
    :param current_vec: StatServer of  vectors not YET normalized
    :param scores: Scores object with symetric matrix of PLDA scores inside
    :param threshold: the clustering threshold
    :param user: user simulation developed in the ALLIES package
    :param clustering_method: complete
    :param selection_method: The method of segment selection for asking quastion from human (could be different for seperation and clustering)
    :param process_all_node: The new stopping creteria for processing whole tree. It changes the limitation of having one line as threshold and lets to consider all branches
    :param conditional_questioning: Usage of reducing std of cluster as condition for asking question
    :return:
    """
#     import pdb
#     pdb.set_trace()
    
    # Perform HAC on the vectors (that shoiuld produce exactly the same segmentation as the one in current diar
    ldiar = copy.deepcopy(current_diar)
    init_diar=copy.deepcopy(initial_diar)
    lscores = copy.deepcopy(scores)
    # get the triangular part of the distances
    distances, th = scores2distance(lscores, threshold)
    distance_sym = squareform(distances)
    
    # t = -1.0 * threshold - min
    # cluster the data
    link = hac.linkage(distance_sym, method=clustering_method)
    
    #code for creation beautiful grafs
    #import pdb
    #pdb.set_trace()
    diar_= copy.deepcopy(initial_diar)
    if False:
        lscores = copy.deepcopy(scores_init)
        scores = copy.deepcopy(scores_init)
        distances, th = scores2distance(lscores, threshold)
        distance_sym = squareform(distances)
        link = hac.linkage(distance_sym, method=clustering_method)
        for i in range (len(diar_)):
            diar_[i]['cluster']=str(i)


    tmp = numpy.zeros((link.shape[0], link.shape[1] + 2))
    tmp[:, :-2] = link
    tmp[:, -2] = link[:, 2] - th
    tmp[:, -1] = numpy.abs(link[:, 2] - th)
    # tmp[:,-2] permet de savoir si on est au dessus du seuil (positif) ou en dessous (négatif)
    # tmp[:, -1] indique la distance au seuil en valeur absolue

    # On trie les regroupements par ordre de proximité au seuil
    # plus proche en premiere ligne
    links_to_check = tmp[numpy.argsort(tmp[:, -1])]

    # On récupère la liste des fusions à faire d'abord.
    final_links = []
    for l in link:
        if l[2] < th:
            final_links.append(l)

    # Maintenant on analyse noeud par noeud du plus proche au plus lointain du seuil de fusion
    # et on gère les deux directions
#     no_more_clustering = False
#     no_more_separation = False
    
    if process_all_nodes:
        separated_list =[]
        separation_list = [] 
        clustering_list = []
    number_cluster = len(scores.scoremat)
    complete_list = list(scores.modelset)
    
    #log the DER
    link_tmp = copy.deepcopy(final_links)
    new_diar_tmp = copy.deepcopy(init_diar)
    cluster_list_tmp = scores.modelset.tolist()
    der, time = check_der(new_diar_tmp,cluster_list_tmp,link_tmp)
    print("Initial DER : ",der)
    
    der_track={"time":time,"der_log":[der],"correction":["initial"],"numQ_clustering":0,"numQ_seperation":0}
    
    for ltc in links_to_check:
        
        # Si on cesse dans les deux directions on sort de la boucle de correction
        if no_more_clustering and no_more_separation:
            break

        elif ltc[-2] < 0:  # On est en dessous du seuil, on a déjà regroupé
            # demande si on sépare: trouve les numéros de deux segments
            # représentatifs des deux cluster déjà regroupés et pose la question
            skip = False
            if process_all_nodes:
                if no_more_separation or set(get_node_spkeakers(ltc[0],number_cluster,link)+get_node_spkeakers(ltc[0],number_cluster,link)).issubset(set(separation_list)) :  # On laisse ensemble
                    skip = True
            else: 
                if no_more_separation:
                    skip = True
            if skip or not check_std_change(ltc, scores, init_diar,current_vec,link,"separate",active=conditional_questioning):
                pass
            else:  # On sépare, concretement on retire la fusion de la liste final_link
                is_same_speaker,_ = ask_question(ltc, scores,scores_init, init_diar,current_vec, user, file_info,link,selection_criteria=selection_method)
                der_track["numQ_seperation"]+=1
                if is_same_speaker:
                    der_track["der_log"].append(der)
                    der_track["correction"].append("None")
                    
                        
#                     print("stop separation ...")
                    if process_all_nodes:
                        separation_list+= get_node_spkeakers(ltc[0],number_cluster,link)
                        separation_list+= get_node_spkeakers(ltc[1],number_cluster,link)
                        if set(complete_list).issubset(set(separation_list)):
                            no_more_separation = True
                    else:
                        no_more_separation = True
    
                       
                else:
                    if process_all_nodes:
                        separated_list+= get_node_spkeakers(ltc[0],number_cluster,link)
                        separated_list+= get_node_spkeakers(ltc[1],number_cluster,link)
#                     print("separate cluster ...")
                    for ii, fl in enumerate(final_links):
                        if numpy.array_equal(fl, ltc[:4]):
                            _ = final_links.pop(ii)
                    link_tmp = copy.deepcopy(final_links)
                    new_diar_tmp = copy.deepcopy(init_diar)
                    cluster_list_tmp = scores.modelset.tolist()
                    der, time = check_der(new_diar_tmp,cluster_list_tmp,link_tmp)
                    der_track["der_log"].append(der)
                    der_track["correction"].append("separate")
                    print("Separate a cluster. new DER : ",der)
                    
                         

        elif ltc[-2] > 0:  # On est au dessous du seuil, on n'a pas encore regroupé
            skip = False
            if process_all_nodes:
                if no_more_clustering or set(get_node_spkeakers(ltc[0],number_cluster,link)+get_node_spkeakers(ltc[0],number_cluster,link)).issubset(set(clustering_list)) :
                    skip = True
                if not set(get_node_spkeakers(ltc[0],number_cluster,link)+get_node_spkeakers(ltc[0],number_cluster,link)).isdisjoint(set(separated_list)):
                    skip = True 
            else:
                if no_more_clustering:
                    skip = True
            if skip or not check_std_change(ltc, scores, init_diar,current_vec,link,"merge",active=conditional_questioning):
                pass

            else:
                is_same_speaker,_ = ask_question(ltc, scores,scores_init, init_diar,current_vec, user, file_info,link,selection_criteria=selection_method)  # true
                der_track["numQ_clustering"]+=1
                if is_same_speaker:  # On fait un nouveau regroupement
#                     print("group clusters ...")
                    final_links.append(ltc[:4])
                    link_tmp = copy.deepcopy(final_links)
                    new_diar_tmp = copy.deepcopy(init_diar)
                    cluster_list_tmp = scores.modelset.tolist()
                    der, time = check_der(new_diar_tmp,cluster_list_tmp,link_tmp)
                    der_track["der_log"].append(der)
                    der_track["correction"].append("cluster")
                    print("Group two clusters. new DER : ",der)
                    
                    
                else:
                    der_track["der_log"].append(der)
                    der_track["correction"].append("None")
                    
                    # On arrête de chercher de ce côté
#                     print("stop grouping ...")
                    if process_all_nodes:
                        clustering_list+= get_node_spkeakers(ltc[0],number_cluster,link)
                        clustering_list+= get_node_spkeakers(ltc[1],number_cluster,link)
                        if set(complete_list).issubset(set(clustering_list)):
                            no_more_clustering = True
                    else:
                        no_more_clustering= True
    
    baselink = copy.deepcopy(link)    
    
    #log the DER
    link = copy.deepcopy(final_links)
    new_diar = copy.deepcopy(init_diar)
    cluster_list = scores.modelset.tolist()
    der, time = check_der(new_diar,cluster_list,link)
    
    print("After corrections, final DER :",der)

    #import pdb
    #pdb.set_trace()

    return False, new_diar, current_vec,der_track



def apply_correction(hyp, scores, c1, c2, method="min"):
    """

    """
    if method == "min":
        hyp = apply_correction_min(hyp, scores, c1, c2)
    elif method == "max":
        hyp = apply_correction_max(hyp, scores, c1, c2)
    elif method == "average":
        hyp = apply_correction_avr(hyp, scores, c1, c2)

    return hyp


def apply_correction_min(hyp,scores,c1,c2):
    c1list=[]
    c2list=[]

    for i in range(len(hyp)):
        if i!=c1 and hyp[i]['cluster'] == hyp[c1]['cluster']:
            c1list.append(i)
        if i!=c2 and hyp[i]['cluster'] == hyp[c2]['cluster']:
            c2list.append(i)

    if len(c1list)==0 and len(c2list)!=0:
        c2_to_c1 = True

    elif len(c2list)==0 and len(c1list)!=0:
        c2_to_c1 = False

    elif len(c1list)==0 and len(c2list)==0:
        c2_to_c1 = True

    else:
        c1dist = []
        c2dist = []
        for i in range(len(c1list)):
            c1dist.append(scores.scoremat[c1][c1list[i]])
        for i in range(len(c2list)):
            c2dist.append(scores.scoremat[c2][c2list[i]])
        if min(c2dist) <  min(c1dist):
            c2_to_c1 = True
        else:
            c2_to_c1 = False

    if c2_to_c1:
        hyp[c1]['cluster'] = hyp[c2]['cluster']
    else:
        hyp[c2]['cluster'] = hyp[c1]['cluster']
    return hyp




def apply_correction_max(hyp,scores,c1,c2):
    c1list=[]
    c2list=[]

    for i in range(len(hyp)):
        if i!=c1 and hyp[i]['cluster'] == hyp[c1]['cluster']:
            c1list.append(i)
        if i!=c2 and hyp[i]['cluster'] == hyp[c2]['cluster']:
            c2list.append(i)

    if len(c1list)==0 and len(c2list)!=0:
        c2_to_c1 = True

    elif len(c2list)==0 and len(c1list)!=0:
        c2_to_c1 = False

    elif len(c1list)==0 and len(c2list)==0:
        c2_to_c1 = True

    else:
        c1dist = []
        c2dist = []
        for i in range(len(c1list)):
            c1dist.append(scores.scoremat[c1][c1list[i]])
        for i in range(len(c2list)):
            c2dist.append(scores.scoremat[c2][c2list[i]])
        if max(c2dist) <  max(c1dist):
            c2_to_c1 = True
        else:
            c2_to_c1 = False

    if c2_to_c1:
        hyp[c1]['cluster'] = hyp[c2]['cluster']
    else:
        hyp[c2]['cluster'] = hyp[c1]['cluster']
    return hyp

def apply_correction_avr(hyp,scores,c1,c2):
    c1list=[]
    c2list=[]

    for i in range(len(hyp)):
        if i!=c1 and hyp[i]['cluster'] == hyp[c1]['cluster']:
            c1list.append(i)
        if i!=c2 and hyp[i]['cluster'] == hyp[c2]['cluster']:
            c2list.append(i)

    if len(c1list)==0 and len(c2list)!=0:
        c2_to_c1 = True

    elif len(c2list)==0 and len(c1list)!=0:
        c2_to_c1 = False

    elif len(c1list)==0 and len(c2list)==0:
        c2_to_c1 = True

    else:
        c1dist = []
        c2dist = []
        for i in range(len(c1list)):
            c1dist.append(scores.scoremat[c1][c1list[i]])
        for i in range(len(c2list)):
            c2dist.append(scores.scoremat[c2][c2list[i]])
        if sum(c2dist)/float(len(c2dist)) <  sum(c1dist)/float(len(c1dist)):
            c2_to_c1 = True
        else:
            c2_to_c1 = False

    if c2_to_c1:
        hyp[c1]['cluster'] = hyp[c2]['cluster']
    else:
        hyp[c2]['cluster'] = hyp[c1]['cluster']
    return hyp




# -* coding: utf-8 -*-

#################################################################################
# This file is part of EVALLIES.                                                #
#                                                                               #
# EVALLIES is a python package for lifelong learning speaker diarization.       #
# Home page: https://git-lium.univ-lemans.fr/Larcher/evallies                   #
#                                                                               #
# EVALLIES is free software: you can redistribute it and/or modify              #
# it under the terms of the GNU LLesser General Public License as               #
# published by the Free Software Foundation, either version 3 of the License,   #
# or (at your option) any later version.                                        #
#                                                                               #
# EVALLIES is distributed in the hope that it will be useful,                   #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 #
# GNU Lesser General Public License for more details.                           #
#                                                                               #
# You should have received a copy of the GNU Lesser General Public License      #
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.              #
#                                                                               #
#################################################################################
"""
Copyright 2020-2021 Anthony Larcher & Olivier Galibert

    :mod:`der_run`

"""

import numpy as np
from scipy.optimize import linear_sum_assignment
import io
from collections import namedtuple
import sys
import os
import os.path


def make_spkmap(spk):
    spkmap = {}
    spkcount = 0
    spkunmap = []
    for s in spk.speaker:
        if not s in spkmap:
            spkunmap.append(s)
            spkmap[s] = spkcount
            spkcount += 1
    return spkmap, spkunmap, spkcount


def range_to_frontiers(rng):
    rng.sort()
    pos = 0
    while pos < len(rng) - 1:
        if rng[pos][1] >= rng[pos + 1][0]:
            rng[pos] = (rng[pos][0], max(rng[pos][1], rng[pos + 1][1]))
            rng.pop(pos + 1)
        else:
            pos = pos + 1
    front = []
    for r in rng:
        front.append(("n", r[0]))
        front.append(("p", r[1]))
    return front

def filter_frontier_on_uem(front, uem):
    uemi = 0
    fri = 0
    fo = []
    while uemi != len(uem.start_time) and fri != len(front):
        if uem.start_time[uemi] < front[fri][1]:
            if uem.end_time[uemi] >= front[fri][1]:
                if front[fri][0] != 'n':
                    if len(fo) == 0 or fo[-1][1] < uem.start_time[uemi]:
                        fo.append(('n', uem.start_time[uemi]))
                    fo.append((front[fri][0], front[fri][1]))
                else:
                    fo.append((front[fri][0], front[fri][1]))
                fri += 1
            else:
                if front[fri][0] != 'n':
                    if len(fo) == 0 or fo[-1][1] < uem.start_time[uemi]:
                        fo.append(('n', uem.start_time[uemi]))
                    fo.append((front[fri][0], uem.end_time[uemi]))
                uemi += 1
        else:
            fri += 1
    return fo

def filter_frontiers_on_uem(front, uem):
    fo = []
    for fr in front:
        fo.append(filter_frontier_on_uem(fr, uem))
    return fo

def merge_two_frontiers(front1, front2, end1, end2):
    frontr = []
    pos1 = 0
    pos2 = 0
    while pos1 < len(front1) or pos2 < len(front2):
        ctime = (
            front1[pos1][1]
            if pos2 == len(front2)
            else front2[pos2][1]
            if pos1 == len(front1)
            else min(front1[pos1][1], front2[pos2][1])
        )
        mode1 = end1 if pos1 == len(front1) else front1[pos1][0]
        mode2 = end2 if pos2 == len(front2) else front2[pos2][0]
        frontr.append((mode1 + mode2, ctime))
        if pos1 != len(front1) and front1[pos1][1] == ctime:
            pos1 += 1
        if pos2 != len(front2) and front2[pos2][1] == ctime:
            pos2 += 1
    return frontr


def make_merge_frontier(hyp_union, ref_union, ref_frontiers_collar):
    hr = merge_two_frontiers(hyp_union, ref_union, "n", "n")
    frontr = []
    for f in ref_frontiers_collar:
        frontr.append(merge_two_frontiers(hr, f, "nn", "n"))
    return frontr


def make_frontiers(spk, spkmap, spkcount):
    rngs = [[] for i in range(spkcount)]
    for i in range(0, len(spk.speaker)):
        spki = spkmap[spk.speaker[i]]
        rngs[spki].append((spk.start_time[i], spk.end_time[i]))
    front = []
    for r in rngs:
        front.append(range_to_frontiers(r))
    return front


def make_union_frontiers(spk):
    rngs = []
    for i in range(0, len(spk.speaker)):
        rngs.append((spk.start_time[i], spk.end_time[i]))
    return range_to_frontiers(rngs)


def frontiers_add_collar(front, collar):
    cfront = []
    for f in front:
        a = f[1] - collar
        b = f[1] + collar
        if a < 0:
            a = 0
        if len(cfront) == 0 or a > cfront[-1][1]:
            cfront.append((f[0], a))
            cfront.append(("t", b))
        else:
            cfront[-1] = ("t", b)
    return cfront


def make_times(front):
    times = []
    for s in front:
        time = 0
        ptime = 0
        for p in s:
            if p[0] == "n":
                ptime = p[1]
            elif p[0] == "p":
                time += p[1] - ptime
        times.append(time)
    return times


def add_time(thyp, thyn, mode, eh, er, tc, efa, emiss, econf):
    if mode == "ppp":
        return eh, er + thyn, tc + thyp, efa, emiss, econf + thyn
    if mode == "ppn":
        return eh + thyp, er, tc, efa, emiss, econf
    if mode == "ppt":
        return eh, er, tc + thyp, efa, emiss, econf
    if mode == "pnn":
        return eh + thyp, er, tc, efa + thyp, emiss, econf
    if mode == "pnt":
        return eh, er, tc + thyp, efa, emiss, econf
    if mode == "npp":
        return eh, er + thyn, tc, efa, emiss + thyn, econf
    # npn npt nnn nnt
    return eh, er, tc, efa, emiss, econf


def compute_times(frontr, fronth):
    eh = 0
    er = 0
    rc = 0
    efa = 0
    emiss = 0
    econf = 0
    hpos = 0
    tbeg = 0
    thyp = 0
    hypbef = 0
    for f in frontr:
        tend = f[1]
        while hpos < len(fronth):
            dinter = min(fronth[hpos][1], tend)
            if fronth[hpos][0] == "p":
                thyp += dinter - hypbef
            if fronth[hpos][1] > tend:
                break
            hypbef = dinter
            hpos += 1
        eh, er, rc, efa, emiss, econf = add_time(
            thyp, tend - tbeg - thyp, f[0], eh, er, rc, efa, emiss, econf
        )

        if hpos < len(fronth):
            hypbef = min(fronth[hpos][1], tend)
        tbeg = tend
        thyp = 0
    while hpos < len(fronth):
        if fronth[hpos][0] == "p":
            thyp += fronth[hpos][1] - tbeg
        tbeg = fronth[hpos][1]
        hpos += 1
    eh, er, rc, efa, emiss, econf = add_time(
        thyp, 0, "pnn", eh, er, rc, efa, emiss, econf
    )
    return (
        round(eh, 3),
        round(er, 3),
        round(rc, 3),
        round(efa, 3),
        round(emiss, 3),
        round(econf, 3),
    )


def compute_miss(funion, front):
    miss = []
    for f1 in front:
        hpos = 0
        tbeg = 0
        thyp = 0
        hypbef = 0
        fa = 0
        for f in funion:
            tend = f[1]
            while hpos < len(f1):
                dinter = min(f1[hpos][1], tend)
                if f1[hpos][0] == "p":
                    thyp += dinter - hypbef
                if f1[hpos][1] > tend:
                    break
                hypbef = dinter
                hpos += 1
            if f[0] == "n":
                fa += thyp
            if hpos < len(f1):
                hypbef = min(f1[hpos][1], tend)
            tbeg = tend
            thyp = 0
        while hpos < len(f1):
            if f1[hpos][0] == "p":
                thyp += f1[hpos][1] - tbeg
            tbeg = f1[hpos][1]
            hpos += 1
        fa += thyp
        fa = round(fa, 3)
        miss.append(fa)
    return miss


def accumulate_confusion(fref, fhyp, map_rh, map_hr):
    ref_spkcount = len(fref)
    hyp_spkcount = len(fhyp)
    correct_ref = [0] * ref_spkcount
    correct_hyp = [0] * hyp_spkcount
    lost_ref = [0] * ref_spkcount
    lost_hyp = [0] * hyp_spkcount
    confusion_matrix = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    fri = [0] * ref_spkcount
    fhi = [0] * hyp_spkcount
    cur_time = 0
    while True:
        ridx = []
        r_is_t = []
        hidx = []
        time = -1

        # Build the list of who is in the segment
        for i in range(ref_spkcount):
            if fri[i] != len(fref[i]):
                cf = fref[i][fri[i]]
                if time == -1 or cf[1] < time:
                    time = cf[1]
                if cf[0] != "n":
                    ridx.append(i)
                    r_is_t.append(cf[0] == "t")

        for i in range(hyp_spkcount):
            if fhi[i] != len(fhyp[i]):
                cf = fhyp[i][fhi[i]]
                if time == -1 or cf[1] < time:
                    time = cf[1]
                if cf[0] != "n":
                    hidx.append(i)

        if time == -1:
            break

        # Only do the computations when there's something to do
        if len(ridx) > 0 or len(hidx) > 0:
            duration = time - cur_time

            # Hyp and ref mapped together end up in correct time and are removed from the lists
            i = 0
            while i != len(ridx):
                r = ridx[i]
                h = map_rh[r]
                dropped = False
                if h != -1:
                    slot = -1
                    for j in range(len(hidx)):
                        if hidx[j] == h:
                            slot = j
                            break
                    if slot != -1:
                        correct_ref[r] += duration
                        correct_hyp[h] += duration
                        ridx.pop(i)
                        r_is_t.pop(i)
                        hidx.pop(slot)
                        dropped = True
                if not dropped:
                    i += 1

            # Ref in transition is removed from the list if mapped to some hyp
            i = 0
            while i != len(ridx):
                r = ridx[i]
                if r_is_t[i] and map_rh[r] != -1:
                    ridx.pop(i)
                    r_is_t.pop(i)
                else:
                    i += 1

            if len(hidx) == 0:
                # If there's no hyp, we're all in lost_ref
                for r in ridx:
                    lost_ref[r] += duration

            elif len(ridx) == 0:
                # If there's no ref, we're all in lost_hyp
                for h in hidx:
                    lost_hyp[h] += duration

            else:
                # Otherwise we're in confusion.  Amount of confusion time to give
                # is equal to the max of the ref and hyp times
                conf_time = max(len(ridx), len(hidx)) * duration

                # Number of slots, otoh, is equal to the product of the number of
                # refs and hyps
                conf_slots = len(ridx) * len(hidx)

                # Give the time equally in all slots
                conf_one_time = conf_time / conf_slots
                for r in ridx:
                    for h in hidx:
                        confusion_matrix[r, h] += conf_one_time

        # Step all the done segments
        for r in range(ref_spkcount):
            if fri[r] != len(fref[r]) and fref[r][fri[r]][1] == time:
                fri[r] += 1
        for h in range(hyp_spkcount):
            if fhi[h] != len(fhyp[h]) and fhyp[h][fhi[h]][1] == time:
                fhi[h] += 1
        cur_time = time

    return correct_ref, correct_hyp, lost_ref, lost_hyp, confusion_matrix


def compute_der(ref, hyp, uem, map, collar):
    ref_spkmap, ref_spkunmap, ref_spkcount = make_spkmap(ref)
    hyp_spkmap, hyp_spkunmap, hyp_spkcount = make_spkmap(hyp)

    ref_frontiers = filter_frontiers_on_uem(make_frontiers(ref, ref_spkmap, ref_spkcount), uem)
    hyp_frontiers = filter_frontiers_on_uem(make_frontiers(hyp, hyp_spkmap, hyp_spkcount), uem)
    ref_frontiers_collar = []
    for front in ref_frontiers:
        ref_frontiers_collar.append(filter_frontier_on_uem(frontiers_add_collar(front, collar), uem))

    ref_union = filter_frontier_on_uem(make_union_frontiers(ref), uem)
    hyp_union = filter_frontier_on_uem(make_union_frontiers(hyp), uem)

    merge_frontiers = make_merge_frontier(hyp_union, ref_union, ref_frontiers_collar)

    ref_times = make_times(ref_frontiers)
    hyp_times = make_times(hyp_frontiers)

    eh = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    er = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    tc = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    efa = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    emiss = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    econf = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    de = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")

    miss_hyp = compute_miss(ref_union, hyp_frontiers)
    miss_ref = compute_miss(hyp_union, ref_frontiers)
    
    for r in range(ref_spkcount):
        for h in range(hyp_spkcount):
            (
                eh[r, h],
                er[r, h],
                tc[r, h],
                efa[r, h],
                emiss[r, h],
                econf[r, h],
            ) = compute_times(merge_frontiers[r], hyp_frontiers[h])
            de[r, h] = ref_times[r] + miss_hyp[h] - efa[r, h] - emiss[r, h] - econf[r, h]

    newspkmap = {}
    map_rh = [-1] * ref_spkcount
    map_hr = [-1] * hyp_spkcount
    solved_ref = [0] * ref_spkcount
    solved_hyp = [0] * hyp_spkcount
    for rs in map:
        hs = map[rs]
        newspkmap[rs] = hs
        rid = ref_spkmap[rs] if rs in ref_spkmap else -1
        hid = hyp_spkmap[hs] if hs in hyp_spkmap else -1
        if rid != -1:
            solved_ref[rid] = -1
        if hid != -1:
            solved_hyp[hid] = -1
        if rid != -1 and hid != -1:
            map_rh[rid] = hid
            map_hr[hid] = rid

    solved_ref_back = []
    solved_hyp_back = []
    for i in range(0, ref_spkcount):
        if solved_ref[i] == 0:
            solved_ref[i] = len(solved_ref_back)
            solved_ref_back.append(i)
    for i in range(0, hyp_spkcount):
        if solved_hyp[i] == 0:
            solved_hyp[i] = len(solved_hyp_back)
            solved_hyp_back.append(i)

    solved_ref_count = len(solved_ref_back)
    solved_hyp_count = len(solved_hyp_back)
    opt_size = max(solved_ref_count, solved_hyp_count)
    costs = np.zeros((opt_size, opt_size), dtype="float64")
    for r1 in range(solved_ref_count):
        r = solved_ref_back[r1]
        for h1 in range(solved_hyp_count):
            h = solved_hyp_back[h1]
            costs[r1, h1] = -round(de[r, h] * 1000)            
        
    (map1, map2) = linear_sum_assignment(costs)
    for i1 in range(0, opt_size):
        i = map1[i1]
        j = map2[i1]
        if (
            i < solved_ref_count
            and j < solved_hyp_count
            and de[i, j] > 0
            and tc[i, j] > 0
        ):
            r = solved_ref_back[i]
            h = solved_hyp_back[j]
            map_rh[r] = h
            map_hr[h] = r

    for i, j in enumerate(map_rh):
        if j != -1:
            newspkmap[ref_spkunmap[i]] = hyp_spkunmap[j]

    ref_mixed_frontiers = []
    for r in range(ref_spkcount):
        if map_rh[r] == -1:
            ref_mixed_frontiers.append(ref_frontiers[r])
        else:
            ref_mixed_frontiers.append(ref_frontiers_collar[r])

    (
        correct_ref,
        correct_hyp,
        lost_ref,
        lost_hyp,
        confusion_matrix,
    ) = accumulate_confusion(ref_mixed_frontiers, hyp_frontiers, map_rh, map_hr)

    conf = 0
    for r in range(ref_spkcount):
        for h in range(hyp_spkcount):
            conf += confusion_matrix[r, h]
    totaltime = 0
    miss = 0
    for r in range(ref_spkcount):
        totaltime += ref_times[r]
        miss += lost_ref[r]
    fa = 0
    for h in range(hyp_spkcount):
        fa += lost_hyp[h]

    return 100 * (fa + miss + conf) / totaltime, totaltime, newspkmap

Spki = namedtuple('Spki', ['speaker', 'start_time', 'end_time'])
Uemi = namedtuple('Uemi', ['start_time', 'end_time'])

def load_mdtm(file):
    global Spki
    res = Spki(speaker=[], start_time=[], end_time=[])
    for l in io.open(file):
        e = l.split()
        res.start_time.append(round(float(e[2]), 3))
        res.end_time.append(round(float(e[2]) + float(e[3]), 3))
        res.speaker.append(e[7])
    return res

def load_uem(file):
    global Uemi
    res = Uemi(start_time=[], end_time=[])
    for l in io.open(file):
        e = l.split()
        res.start_time.append(round(float(e[2]), 3))
        res.end_time.append(round(float(e[3]), 3))
    return res

    
def load_map(file):
    res = {}
    for l in io.open(file):
        e = l.split()
        res[e[0]] = e[1]
    return res

def find_file(pdir, names, ext):
    for i in range(0, 2):
        p = pdir + '/' + names[i] + ext
        if os.path.isfile(p):
            return p
    err = ""
    for i in range(0, 2):
        if i == 0:
            err += "Neither "
        else:
            err += " nor "
        err += pdir + '/' + names[i] + ext
    err += ' war found.'
    print(err, file = sys.stderr)
    sys.exit(1)



def der_incremental(lifelong_list, ref_mdtm_dir, ref_uem_dir, hyp_mdtm_dir):

    map = {}

    total_error = 0
    total_time = 0

    for l in open(lifelong_list):
        e = l.split()
        ref = load_mdtm(find_file(ref_mdtm_dir, e, ".mdtm"))
        uem = load_uem (find_file(ref_uem_dir, e, ".uem"))
        hyp = load_mdtm(find_file(hyp_mdtm_dir, e, ".mdtm"))
        (der, time, map) = compute_der(ref, hyp, uem, map, 0.250)
        total_error += der * time
        total_time += time
        #print("%s: DER: %6.2f%%  Time: %.2fs" % (e[0], der, time))

    print("*** Cross Show Incremental DER: %6.2f%%  Total time: %.2fs" % (total_error / total_time, total_time))

    #mapf = open(sys.argv[5], 'w')
    #for s in map:
    #    print("%s %s" % (s, map[s]), file=mapf)





# -* coding: utf-8 -*-
#
# This file is part of EVALLIES.
#
# EVALLIES is a python package for lifelong learning speaker diarization.
# Home page: https://git-lium.univ-lemans.fr/Larcher/evallies
#
# EVALLIES is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# EVALLIES is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.
"""
Copyright 2020-2021 Anthony Larcher

    :mod:`init`

"""


from .database import Database

from .user_simulation import UserSimulation
from .user_simulation import MessageToUser 
from .user_simulation import Request

from .der_single import der_single_averaged
from .der_run import der_incremental
from .der_cross import der_cross

#from .lium_baseline.system import allies_load_seg
from .lium_baseline.system import allies_initial_training
from .lium_baseline.system import lium_iv_initial_training
from .lium_baseline.system import lium_xv_initial_training
from .lium_baseline.system import extract_vectors
from .lium_baseline.system import allies_init_seg
from .lium_baseline.system import allies_within_show_HAL
from .lium_baseline.system import allies_cross_show_clustering
from .lium_baseline.system import allies_write_diar

from .lium_baseline.interactive import apply_correction
from .lium_baseline.interactive import apply_correction_min
from .lium_baseline.interactive import apply_correction_max
from .lium_baseline.interactive import apply_correction_avr

__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2020-2021 Anthony Larcher"
__license__ = "LGPL"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'
__version__="0.1.0"



# -* coding: utf-8 -*-

#################################################################################
# This file is part of EVALLIES.                                                #
#                                                                               #
# EVALLIES is a python package for lifelong learning speaker diarization.       #
# Home page: https://git-lium.univ-lemans.fr/Larcher/evallies                   #
#                                                                               #
# EVALLIES is free software: you can redistribute it and/or modify              #
# it under the terms of the GNU LLesser General Public License as               #
# published by the Free Software Foundation, either version 3 of the License,   #
# or (at your option) any later version.                                        #
#                                                                               #
# EVALLIES is distributed in the hope that it will be useful,                   #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 #
# GNU Lesser General Public License for more details.                           #
#                                                                               #
# You should have received a copy of the GNU Lesser General Public License      #
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.              #
#                                                                               #
#################################################################################
"""
Copyright 2020-2021 Anthony Larcher, Meysam Shamsi & Yevhenii Propkopalo

    :mod:`lium_baseline.utils`

"""


import copy
import numpy
import sidekit
from ..user_simulation import Reference

def add_show_in_id(im):
    """
    Work on StatServer 
    """
    im2 = copy.deepcopy(im)
    max_length = 0
    for mod, seg in zip(im.modelset, im.segset):
        if max_length < 1 + len(mod) + len(seg):
            max_length = 1 + len(mod) + len(seg)
            im2.modelset = im2.modelset.astype('<U{}'.format(max_length))

    for ii, c_id in enumerate(im.modelset):
        im2.modelset[ii] = im.segset[ii] + "$" + c_id

    return im2


def remove_show_from_id(im):
    """
    Work on StatServer
    :param im:
    :return:
    """
    im2 = copy.deepcopy(im)
    im2.modelset = im2.modelset.astype('U', copy=False)
    for idx, c_id in enumerate(im.modelset):
        im2.modelset[idx] = c_id.split('$')[1]
    return im2

def s4d_to_allies(s4d_segmentation):
    """
    Convert a s4d.Diar object into the proper ALLIES format for segmentation
    :param s4d_segmentation: is a s4d.Diar object
    """
    spk = []
    st = []
    en = []
    for segment in s4d_segmentation:
        spk.append(segment["cluster"])
        st.append(float(segment["start"]) / 100.)
        en.append(float(segment["stop"]) / 100.)
    hyp = Reference(spk, st, en)
    #hyp = {"speaker": spk, "start_time": st, "end_time": en}
    return hyp

def keep_recurring_speakers(iv_ss, rank_F, occ_number=2, filter_by_name=False):
    """
    """
    unique, counts = numpy.unique(iv_ss.modelset, return_counts=True)
    tot_sessions = 0
    kept_idx = []
    for model in range(unique.shape[0]):
        if counts[model] >= occ_number and (
            (
                not (unique[model]).startswith("speaker")
                and not (unique[model]).startswith("presentat")
                and not (unique[model]).startswith("sup+")
                and not (unique[model]).startswith("voix")
                and not (unique[model]).startswith("publicitaire")
                and not (unique[model]).startswith("locuteur")
                and not (unique[model]).startswith("inconnu")
                and not (unique[model]).startswith("traduct")
                and not ("#" in (unique[model]))
                and "_" in (unique[model])
                and filter_by_name
            )
            or (not filter_by_name)
        ):
            kept_idx = numpy.append(
                kept_idx, numpy.where(iv_ss.modelset == unique[model])
            )
            tot_sessions += counts[model]
    #
    flt_iv = sidekit.StatServer()
    flt_iv.stat0 = numpy.ones((tot_sessions, 1))
    flt_iv.stat1 = numpy.ones((tot_sessions, rank_F))
    flt_iv.modelset = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.segset = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.start = numpy.empty(tot_sessions, dtype="|O")
    flt_iv.stop = numpy.empty(tot_sessions, dtype="|O")
    #
    i = 0
    for model in range(unique.shape[0]):
        if counts[model] >= occ_number and (
            (
                not (unique[model]).startswith("speaker")
                and not (unique[model]).startswith("presentat")
                and not (unique[model]).startswith("sup+")
                and not (unique[model]).startswith("voix")
                and not (unique[model]).startswith("publicitaire")
                and not (unique[model]).startswith("locuteur")
                and not (unique[model]).startswith("inconnu")
                and not (unique[model]).startswith("traduct")
                and not ("#" in (unique[model]))
                and "_" in (unique[model])
                and filter_by_name
            )
            or (not filter_by_name)
        ):
            for ivector in numpy.where(iv_ss.modelset == unique[model])[0]:
                flt_iv.stat1[i, :] = iv_ss.stat1[ivector, :]
                flt_iv.modelset[i] = iv_ss.modelset[ivector]
                flt_iv.segset[i] = iv_ss.segset[ivector]
                flt_iv.start[i] = iv_ss.start[ivector]
                flt_iv.stop[i] = iv_ss.stop[ivector]
                i += 1
    unique = numpy.unique(flt_iv.modelset)
    return flt_iv, unique.shape[0], kept_idx





# -* coding: utf-8 -*-

#################################################################################
# This file is part of EVALLIES.                                                #
#                                                                               #
# EVALLIES is a python package for lifelong learning speaker diarization.       #
# Home page: https://git-lium.univ-lemans.fr/Larcher/evallies                   #
#                                                                               #
# EVALLIES is free software: you can redistribute it and/or modify              #
# it under the terms of the GNU LLesser General Public License as               #
# published by the Free Software Foundation, either version 3 of the License,   #
# or (at your option) any later version.                                        #
#                                                                               #
# EVALLIES is distributed in the hope that it will be useful,                   #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 #
# GNU Lesser General Public License for more details.                           #
#                                                                               #
# You should have received a copy of the GNU Lesser General Public License      #
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.              #
#                                                                               #
#################################################################################
"""
Copyright 2020-2021 Anthony Larcher, Olivier Galibert

    :mod:`database`

"""

import os
import numpy as np
import wave
import struct
from collections import namedtuple
from .user_simulation import FileInfo
from .user_simulation import Reference
from .user_simulation import UEM

def get_key(filename):
    if filename[0:3] == "199" or filename[0:3] == "200":
        return filename[0:8] + "." + filename[9:13] + "." + filename[19:]
    l = len(filename)
    if filename[l - 17 : l - 14] == "201":
        return (
            filename[l - 17 : l - 13]
            + filename[l - 12 : l - 10]
            + filename[l - 9 : l - 7]
            + "."
            + filename[l - 6 : l - 2]
            + "."
            + filename[0 : l - 18]
        )
    if filename[14:18] == "2010":
        if filename[27:28] == "_":
            return filename[14:22] + "." + filename[23:27] + "." + filename[28:]
        else:
            return filename[14:22] + "." + filename[23:27] + "." + filename[30:]
    if filename[:3] == "201":
        pp = filename.split(".")[0].split("_")
        return pp[0] + "." + "_".join(pp[1:-1]) + "." + pp[-1]
    return "???"


def get_file_list(path):
    pairs = []
    for name in os.listdir(path):
        if not name.startswith('.'):
            name = name[:-5]
            k = get_key(name)
            pairs.append([k, name])
    pairs.sort(key=lambda p: p[0])
    return pairs


def hashrand(a):
    a = (a + 0x7ED55D16) + (a << 12)
    a = (a ^ 0xC761C23C) ^ (a >> 19)
    a = (a + 0x165667B1) + (a << 5)
    a = (a + 0xD3A2646C) ^ (a << 9)
    a = (a + 0xFD7046C5) + (a << 3)
    a = (a ^ 0xB55A4F09) ^ (a >> 16)
    a = a & 0xFFFFFFFF
    a = a / 0xFFFFFFFF
    return a


class Database:
    """
    Object that load data from the ALLIES database and create an iterator that will
    yield requested information to process a file:
    file_info, uem, speakers

    speakers is the reference from MDTM

    Can select according to the TV> channel : BFM or LCP and depending on the show:

    """
    def __init__(self, root_folder, parameters, show=None):
        """

        :param root_folder:
        :param parameters:
        :param show:    can be LCP, BFM,  topquestions, pileetface, cultureetvous, cavousregarde, lcpinfo, entre, ruthelkrief
        """


        self.index = 0
        self.root_folder = root_folder

        self.fl = get_file_list(root_folder + "/mdtm")

        Entry = namedtuple(
            # 'Entry', ['root_folder', 'filename', 'file_info', 'speech', 'uem', 'speakers', 'llmode']
            "Entry",
            ["root_folder", "filename", "file_info", "speech", "uem", "speakers"],
        )

        self.ents = []
        for idx, f in enumerate(self.fl):

            # Select according to the show of channel
            if show is None or show in f[0].lower():
                if ("start_date" in parameters) and f[0] < parameters["start_date"]:
                    continue
                if ("end_date" in parameters) and f[0] >= parameters["end_date"]:
                    continue
                llmode = "none"
                if "interactive_ratio" in parameters:
                    if float(parameters["interactive_ratio"]) > hashrand(idx):
                        llmode = "interactive"
                    else:
                        llmode = "active"
                self.ents.append(
                    Entry(
                        root_folder,
                        f[1],
                        FileInfo(f[0], llmode, f[0][0:13]),
                        [f[0]],
                        [f[0]],
                        [f[0]],
                    )
                )
        self.number_of_files = len(self.ents)

    def __iter__(self):
        return self

    def __next__(self):
        """
        Return a tuple including 
            file_info, uem, speakers
        :return: 
        """
        if self.index < self.number_of_files:

            # GET filename
            filename = self.ents[self.index].filename
            complete_filename = self.root_folder + "/wav/" + self.ents[self.index].filename + ".wav"

            # GET file_info
            file_info = self.ents[self.index].file_info

            # GET UEM
            st = []
            en = []
            for l in open(self.root_folder + "/uem/" + self.ents[self.index].filename + ".uem", "r"):
                e = l.split()
                st.append(e[2])
                en.append(e[3])
            #uem = {"start_time": np.cast["float64"](st), "end_time": np.cast["float64"](en)}
            uem = UEM(np.cast["float64"](st), np.cast["float64"](en))

            # GET Speakers
            spk = []
            ref_st = []
            ref_en = []
            for l in open(self.root_folder + "/mdtm/" + self.ents[self.index].filename + ".mdtm", "r"):
                e = l.split()
                ref_st.append(np.cast["float64"](round(float(e[2]), 3)))
                ref_en.append(np.cast["float64"](round(float(e[2]) + float(e[3]), 3)))
                spk.append(e[7])
            speakers = Reference(spk, ref_st, ref_en)
            self.index += 1
            return filename, file_info, uem, speakers, complete_filename

        else:
            raise StopIteration()



# -* coding: utf-8 -*-

#################################################################################
# This file is part of EVALLIES.                                                #
#                                                                               #
# EVALLIES is a python package for lifelong learning speaker diarization.       #
# Home page: https://git-lium.univ-lemans.fr/Larcher/evallies                   #
#                                                                               #
# EVALLIES is free software: you can redistribute it and/or modify              #
# it under the terms of the GNU LLesser General Public License as               #
# published by the Free Software Foundation, either version 3 of the License,   #
# or (at your option) any later version.                                        #
#                                                                               #
# EVALLIES is distributed in the hope that it will be useful,                   #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 #
# GNU Lesser General Public License for more details.                           #
#                                                                               #
# You should have received a copy of the GNU Lesser General Public License      #
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.              #
#                                                                               #
#################################################################################
"""
Copyright 2020-2021 Anthony Larcher & Olivier Galibert

    :mod:`user_simulation`

"""
import numpy as np

from scipy.optimize import linear_sum_assignment


def make_spkmap(spk):
    spkmap = {}
    spkcount = 0
    for s in spk.speaker:
        if not s in spkmap:
            spkmap[s] = spkcount
            spkcount += 1
    return spkmap, spkcount


def range_to_frontiers(rng):
    rng.sort()
    pos = 0
    while pos < len(rng) - 1:
        if rng[pos][1] >= rng[pos + 1][0]:
            rng[pos] = (rng[pos][0], max(rng[pos][1], rng[pos + 1][1]))
            rng.pop(pos + 1)
        else:
            pos = pos + 1
    front = []
    for r in rng:
        front.append(("n", r[0]))
        front.append(("p", r[1]))
    return front


def merge_two_frontiers(front1, front2, end1, end2):
    frontr = []
    pos1 = 0
    pos2 = 0
    while pos1 < len(front1) or pos2 < len(front2):
        ctime = (
            front1[pos1][1]
            if pos2 == len(front2)
            else front2[pos2][1]
            if pos1 == len(front1)
            else min(front1[pos1][1], front2[pos2][1])
        )
        mode1 = end1 if pos1 == len(front1) else front1[pos1][0]
        mode2 = end2 if pos2 == len(front2) else front2[pos2][0]
        frontr.append((mode1 + mode2, ctime))
        if pos1 != len(front1) and front1[pos1][1] == ctime:
            pos1 += 1
        if pos2 != len(front2) and front2[pos2][1] == ctime:
            pos2 += 1
    return frontr


def make_merge_frontier(hyp_union, ref_union, ref_frontiers_collar):
    hr = merge_two_frontiers(hyp_union, ref_union, "n", "n")
    frontr = []
    for f in ref_frontiers_collar:
        frontr.append(merge_two_frontiers(hr, f, "nn", "n"))
    return frontr


def make_frontiers(spk, spkmap, spkcount):
    rngs = [[] for i in range(spkcount)]
    for i in range(0, len(spk.speaker)):
        spki = spkmap[spk.speaker[i]]
        rngs[spki].append((spk.start_time[i], spk.end_time[i]))
    front = []
    for r in rngs:
        front.append(range_to_frontiers(r))
    return front


def make_union_frontiers(spk):
    rngs = []
    for i in range(0, len(spk.speaker)):
        rngs.append((spk.start_time[i], spk.end_time[i]))
    return range_to_frontiers(rngs)


def frontiers_add_collar(front, collar):
    cfront = []
    for f in front:
        a = f[1] - collar
        b = f[1] + collar
        if a < 0:
            a = 0
        if len(cfront) == 0 or a > cfront[-1][1]:
            cfront.append((f[0], a))
            cfront.append(("t", b))
        else:
            cfront[-1] = ("t", b)
    return cfront


def make_times(front):
    times = []
    for s in front:
        time = 0
        ptime = 0
        for p in s:
            if p[0] == "n":
                ptime = p[1]
            elif p[0] == "p":
                time += p[1] - ptime
        times.append(time)
    return times


def add_time(thyp, thyn, mode, eh, er, tc, efa, emiss, econf):
    if mode == "ppp":
        return eh, er + thyn, tc + thyp, efa, emiss, econf + thyn
    if mode == "ppn":
        return eh + thyp, er, tc, efa, emiss, econf
    if mode == "ppt":
        return eh, er, tc + thyp, efa, emiss, econf
    if mode == "pnn":
        return eh + thyp, er, tc, efa + thyp, emiss, econf
    if mode == "pnt":
        return eh, er, tc + thyp, efa, emiss, econf
    if mode == "npp":
        return eh, er + thyn, tc, efa, emiss + thyn, econf
    # npn npt nnn nnt
    return eh, er, tc, efa, emiss, econf


def compute_times(frontr, fronth):
    eh = 0
    er = 0
    rc = 0
    efa = 0
    emiss = 0
    econf = 0
    hpos = 0
    tbeg = 0
    thyp = 0
    hypbef = 0
    for f in frontr:
        tend = f[1]
        while hpos < len(fronth):
            dinter = min(fronth[hpos][1], tend)
            if fronth[hpos][0] == "p":
                thyp += dinter - hypbef
            if fronth[hpos][1] > tend:
                break
            hypbef = dinter
            hpos += 1
        eh, er, rc, efa, emiss, econf = add_time(
            thyp, tend - tbeg - thyp, f[0], eh, er, rc, efa, emiss, econf
        )

        if hpos < len(fronth):
            hypbef = min(fronth[hpos][1], tend)
        tbeg = tend
        thyp = 0
    while hpos < len(fronth):
        if fronth[hpos][0] == "p":
            thyp += fronth[hpos][1] - tbeg
        tbeg = fronth[hpos][1]
        hpos += 1
    eh, er, rc, efa, emiss, econf = add_time(
        thyp, 0, "pnn", eh, er, rc, efa, emiss, econf
    )
    return (
        round(eh, 3),
        round(er, 3),
        round(rc, 3),
        round(efa, 3),
        round(emiss, 3),
        round(econf, 3),
    )


def compute_miss(funion, front):
    miss = []
    for f1 in front:
        hpos = 0
        tbeg = 0
        thyp = 0
        hypbef = 0
        fa = 0
        for f in funion:
            tend = f[1]
            while hpos < len(f1):
                dinter = min(f1[hpos][1], tend)
                if f1[hpos][0] == "p":
                    thyp += dinter - hypbef
                if f1[hpos][1] > tend:
                    break
                hypbef = dinter
                hpos += 1
            if f[0] == "n":
                fa += thyp
            if hpos < len(f1):
                hypbef = min(f1[hpos][1], tend)
            tbeg = tend
            thyp = 0
        while hpos < len(f1):
            if f1[hpos][0] == "p":
                thyp += f1[hpos][1] - tbeg
            tbeg = f1[hpos][1]
            hpos += 1
        fa += thyp
        fa = round(fa, 3)
        miss.append(fa)
    return miss


def accumulate_confusion(fref, fhyp, map_rh, map_hr):
    ref_spkcount = len(fref)
    hyp_spkcount = len(fhyp)
    correct_ref = [0] * ref_spkcount
    correct_hyp = [0] * hyp_spkcount
    lost_ref = [0] * ref_spkcount
    lost_hyp = [0] * hyp_spkcount
    confusion_matrix = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    fri = [0] * ref_spkcount
    fhi = [0] * hyp_spkcount
    cur_time = 0
    while True:
        ridx = []
        r_is_t = []
        hidx = []
        time = -1

        # Build the list of who is in the segment
        for i in range(ref_spkcount):
            if fri[i] != len(fref[i]):
                cf = fref[i][fri[i]]
                if time == -1 or cf[1] < time:
                    time = cf[1]
                if cf[0] != "n":
                    ridx.append(i)
                    r_is_t.append(cf[0] == "t")

        for i in range(hyp_spkcount):
            if fhi[i] != len(fhyp[i]):
                cf = fhyp[i][fhi[i]]
                if time == -1 or cf[1] < time:
                    time = cf[1]
                if cf[0] != "n":
                    hidx.append(i)

        if time == -1:
            break

        # Only do the computations when there's something to do
        if len(ridx) > 0 or len(hidx) > 0:
            duration = time - cur_time

            # Hyp and ref mapped together end up in correct time and are removed from the lists
            i = 0
            while i != len(ridx):
                r = ridx[i]
                h = map_rh[r]
                dropped = False
                if h != -1:
                    slot = -1
                    for j in range(len(hidx)):
                        if hidx[j] == h:
                            slot = j
                            break
                    if slot != -1:
                        correct_ref[r] += duration
                        correct_hyp[h] += duration
                        ridx.pop(i)
                        r_is_t.pop(i)
                        hidx.pop(slot)
                        dropped = True
                if not dropped:
                    i += 1

            # Ref in transition is removed from the list if mapped to some hyp
            i = 0
            while i != len(ridx):
                r = ridx[i]
                if r_is_t[i] and map_rh[r] != -1:
                    ridx.pop(i)
                    r_is_t.pop(i)
                else:
                    i += 1

            if len(hidx) == 0:
                # If there's no hyp, we're all in lost_ref
                for r in ridx:
                    lost_ref[r] += duration

            elif len(ridx) == 0:
                # If there's no ref, we're all in lost_hyp
                for h in hidx:
                    lost_hyp[h] += duration

            else:
                # Otherwise we're in confusion.  Amount of confusion time to give
                # is equal to the max of the ref and hyp times
                conf_time = max(len(ridx), len(hidx)) * duration

                # Number of slots, otoh, is equal to the product of the number of
                # refs and hyps
                conf_slots = len(ridx) * len(hidx)

                # Give the time equally in all slots
                conf_one_time = conf_time / conf_slots
                for r in ridx:
                    for h in hidx:
                        confusion_matrix[r, h] += conf_one_time

        # Step all the done segments
        for r in range(ref_spkcount):
            if fri[r] != len(fref[r]) and fref[r][fri[r]][1] == time:
                fri[r] += 1
        for h in range(hyp_spkcount):
            if fhi[h] != len(fhyp[h]) and fhyp[h][fhi[h]][1] == time:
                fhi[h] += 1
        cur_time = time

    return correct_ref, correct_hyp, lost_ref, lost_hyp, confusion_matrix


def find_common_point(f1, f2):
    fr = merge_two_frontiers(f1, f2, "nn", "n")
    st = None
    en = None
    dur = None
    for i in range(1, len(fr)):
        if fr[i][0] == "pp":
            st1 = fr[i - 1][1]
            en1 = fr[i][1]
            dur1 = en1 - st1
            if dur == None or dur < dur1:
                st = st1
                en = en1
                dur = dur1
    # Should we randomize?  Let's be nice for now
    return (st + en) / 2


def find_best_information(ref, hyp, collar, correction_type=(True, True, True)):
    """

    :param ref:
    :param hyp:
    :param collar:
    :param correction_type: a triplet of boolean. Each value is associated to a type of correction:
        first one is merge two segments, second one: split two segments and third one: modify the frontiers of a segment
    :return:
    """
    ref_spkmap, ref_spkcount = make_spkmap(ref)
    hyp_spkmap, hyp_spkcount = make_spkmap(hyp)

    ref_frontiers = make_frontiers(ref, ref_spkmap, ref_spkcount)
    hyp_frontiers = make_frontiers(hyp, hyp_spkmap, hyp_spkcount)
    ref_frontiers_collar = []
    for front in ref_frontiers:
        ref_frontiers_collar.append(frontiers_add_collar(front, collar))

    ref_union = make_union_frontiers(ref)
    hyp_union = make_union_frontiers(hyp)

    merge_frontiers = make_merge_frontier(hyp_union, ref_union, ref_frontiers_collar)

    ref_times = make_times(ref_frontiers)
    hyp_times = make_times(hyp_frontiers)

    eh = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    er = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    tc = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    efa = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    emiss = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    econf = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")
    de = np.zeros((ref_spkcount, hyp_spkcount), dtype="float64")

    opt_size = max(ref_spkcount, hyp_spkcount)
    costs = np.zeros((opt_size, opt_size), dtype="float64")

    miss_hyp = compute_miss(ref_union, hyp_frontiers)
    miss_ref = compute_miss(hyp_union, ref_frontiers)

    for r in range(ref_spkcount):
        for h in range(hyp_spkcount):
            (
                eh[r, h],
                er[r, h],
                tc[r, h],
                efa[r, h],
                emiss[r, h],
                econf[r, h],
            ) = compute_times(merge_frontiers[r], hyp_frontiers[h])
            de[r, h] = (
                    ref_times[r] + miss_hyp[h] - efa[r, h] - emiss[r, h] - econf[r, h]
            )
            costs[r, h] = -round(de[r, h] * 1000)

    (map1, map2) = linear_sum_assignment(costs)
    map_rh = [-1] * ref_spkcount
    map_hr = [-1] * hyp_spkcount
    for i1 in range(0, opt_size):
        i = map1[i1]
        j = map2[i1]
        if (
                i < ref_spkcount
                and j < hyp_spkcount
                and de[i, j] > 0
                and tc[i, j] > 0
        ):
            map_rh[i] = j
            map_hr[j] = i

    ref_mixed_frontiers = []
    for r in range(ref_spkcount):
        if map_rh[r] == -1:
            ref_mixed_frontiers.append(ref_frontiers[r])
        else:
            ref_mixed_frontiers.append(ref_frontiers_collar[r])

    (
        correct_ref,
        correct_hyp,
        lost_ref,
        lost_hyp,
        confusion_matrix,
    ) = accumulate_confusion(ref_mixed_frontiers, hyp_frontiers, map_rh, map_hr)

    conf = 0
    for r in range(ref_spkcount):
        for h in range(hyp_spkcount):
            conf += confusion_matrix[r, h]
    totaltime = 0
    miss = 0
    for r in range(ref_spkcount):
        totaltime += ref_times[r]
        miss += lost_ref[r]
    fa = 0
    for h in range(hyp_spkcount):
        fa += lost_hyp[h]

    best_segment_fix = None

    # Do one pass over the reference segments to see how much to gain if one hypothesis segment is corrected
    # Assume the speaker is correct if there's only one, or that the biggest is correct otherwise

    nf = len(ref_frontiers[0])
    for i in range(0, len(ref.speaker)):
        # segment boundaries
        st = ref.start_time[i]
        en = ref.end_time[i]

        # amount of silence before and after the segment
        silence_before = 0
        silence_after = 0
        for f in ref_union:
            if f[1] < st:
                silence_before = st - f[1] if f[0] == 'p' else 0
            if f[1] > en:
                silence_after = f[1] - en if f[0] == 'n' else 0
                break

        # scan the hypothesis, collate the per-speaker time
        hyptime = {}
        for j in range(0, len(hyp_frontiers)):
            hst = hyp.start_time[j]
            hen = hyp.end_time[j]
            if not (hen <= st or hst >= en):
                if hen > en:
                    hen = en
                if hst < st:
                    hst = st
                hspk = hyp.speaker[j]
                if hspk in hyptime:
                    hyptime[hspk] += hen - hst
                else:
                    hyptime[hspk] = hen - hst

        # compute the fixed time under the assumption that the longest speaker is correct
        fixed_time = 0
        if len(hyptime) > 1:
            best_time = 0
            for s in hyptime:
                tm = hyptime[s]
                if tm > best_time:
                    best_time = tm
                fixed_time += tm
            fixed_time -= best_time

        # compute the time in speech in the silences on the border, and the time in silence in the speech
        if len(hyp_union) > 0:
            stf = 0
            enf = len(hyp_union) - 1
            for j in range(0, len(hyp_union)):
                if hyp_union[j][1] < st:
                    stf = j
                if hyp_union[j][1] > en:
                    enf = j
                    break
            # pre/post-segment silence
            if hyp_union[stf][1] < st and hyp_union[stf][0] == 'n':
                sp = st - silence_before;
                if sp < hyp_union[stf][1]:
                    sp = hyp_union[stf][1]
                fixed_time += st - sp
            if hyp_union[enf][1] > en and hyp_union[enf][0] == 'p':
                sp = en + silence_after;
                if sp > hyp_union[enf][1]:
                    sp = hyp_union[enf][1]
                fixed_time += sp - en
            # in-segment silence
            sp = st
            for j in range(stf, enf + 1):
                if hyp_union[j][1] > st and hyp_union[j][0] == 'p':
                    sp = hyp_union[j][1]
                if hyp_union[j][1] > st and hyp_union[j][1] < en and hyp_union[j][0] == 'n':
                    fixed_time += hyp_union[j][1] - sp

        else:
            # nothing in the hypothesis, all the time is fixed
            fixed_time += en - st;
        if fixed_time:
            if (best_segment_fix == None) or (best_segment_fix[1] < fixed_time):
                best_segment_fix = [fixed_time, st, en]

    # find the couple of maximum confusion where both sides are mapped
    max_conf = 0
    max_conf_r = None
    max_conf_h = None
    for r in range(ref_spkcount):
        for h in range(hyp_spkcount):
            if confusion_matrix[r, h] > max_conf and map_rh[r] != -1 and map_hr[h] != -1:
                max_conf = confusion_matrix[r, h]
                max_conf_r = r
                max_conf_h = h
    if max_conf_r != None:
        # Of the two, pick the speaker with the maximum amount of error associated
        error_spk_ref = lost_ref[r]
        for h in range(hyp_spkcount):
            error_spk_ref += confusion_matrix[max_conf_r, h]
        error_spk_hyp = lost_hyp[h]
        for r in range(ref_spkcount):
            error_spk_hyp += confusion_matrix[r, max_conf_h]

        if (error_spk_ref > error_spk_hyp) or not correction_type[1]:
            # We want to pivot on the reference, that means merging the mapped hyp speaker and the mapped max error hyp speaker
            correct_point = find_common_point(ref_frontiers[max_conf_r], hyp_frontiers[map_rh[max_conf_r]])
            bad_point = find_common_point(ref_frontiers[max_conf_r], hyp_frontiers[max_conf_h])
            p1 = min(correct_point, bad_point)
            p2 = max(correct_point, bad_point)
            max_conf_a = Answer(True, "same", np.float32(p1), np.float32(p2))
            #max_conf_a = {"answer": {"value": True}, "response_type": "same", "time_1": np.float32(p1),
            #              "time_2": np.float32(p2)}
        elif (error_spk_hyp >= error_spk_ref) or not correction_type[0]:
            # We want to pivot on the hypothesis, that means splitting the mapped ref speaker and the mapped max error ref speaker
            correct_point = find_common_point(ref_frontiers[map_hr[max_conf_h]], hyp_frontiers[max_conf_h])
            bad_point = find_common_point(ref_frontiers[max_conf_r], hyp_frontiers[max_conf_h])
            #print(f"{correct_point} {bad_point}")
            p1 = min(correct_point, bad_point)
            p2 = max(correct_point, bad_point)
            max_conf_a = Answer(False, "same", np.float32(p1), np.float32(p2))
            #max_conf_a = {"answer": {"value": False}, "response_type": "same", "time_1": np.float32(p1),
            #              "time_2": np.float32(p2)}

    if (best_segment_fix == None and max_conf == 0) or (correction_type == (False, False, False)):
        #return {"answer": {"value": False}, "response_type": "stop", "time_1": np.float32(0.0),
        #        "time_2": np.float32(0.0)}
        return Answer(False, "stop", np.float32(0.0), np.float32(0.0))
    if best_segment_fix == None or max_conf > best_segment_fix[0] or not correction_type[2]:
        return max_conf_a
    else:
        #return {"answer": {"value": True}, "response_type": "boundary", "time_1": np.float32(best_segment_fix[1]),
        #        "time_2": np.float32(best_segment_fix[2])}
        return Answer(True, "boundary", np.float32(best_segment_fix[1]), np.float32(best_segment_fix[2]))


class UserSimulation:

    def __init__(self, parameters, correction_type=(True, True, True)):
        self.cost = 0
        self.max_cost_per_file = parameters["max_cost_per_file"]
        self.request_collar_cost = parameters["request_collar_cost"]
        self.correction_type = correction_type


    def compute_answer_cost(self, a):
        cost = 0
        #if a["response_type"] == "same":
        if a.response_type == "same":
            #time_1 = a["time_1"]
            #time_2 = a["time_2"]
            time_1 = a.time_1
            time_2 = a.time_2
            if abs(time_2 - time_1) >= self.request_collar_cost:
                cost += 2 * self.request_collar_cost
            else:
                cost += max(time_1, time_2) - min(time_1, time_2) + self.request_collar_cost
        elif a.response_type == "boundary":
            #elif a["response_type"] == "boundary":
            #time_1 = a["time_1"]
            #time_2 = a["time_2"]
            time_1 = a.time_1
            time_2 = a.time_2
            cost = time_2 - time_1 + self.request_collar_cost
        return cost

    def read(self, file_info, uem, reference):
        self.file_info = file_info
        self.reference = reference
        self.uem = uem

    def validate(self, message):
        answer = {}
#         print("entered validate:")
        #print(f"cost = {self.cost}, et max = {self.max_cost_per_file}")
        if self.cost >= self.max_cost_per_file:
            answer = Answer(False, "stop", 0.0, 0.0)
            #answer = {
            #    "answer": {"value": False},
            #    "response_type": "stop",
            #    "time_1": 0.0,
            #    "time_2": 0.0,
            #}
        elif message.file_info.supervision == "active":
            if message.system_request.request_type == "same":
                time_1 = message.system_request.time_1
                time_2 = message.system_request.time_2
                spk1 = self.find_speaker_for_time(time_1)
                spk2 = self.find_speaker_for_time(time_2)
                if abs(time_2 - time_1) >= self.request_collar_cost:
                    self.cost += 2 * self.request_collar_cost
                else:
                    self.cost += max(time_1, time_2) - min(time_1, time_2) + self.request_collar_cost
                print(
                    "USER: Check for same on %f (%s) vs. %f (%s)"
                    % (time_1, spk1, time_2, spk2)
                )
                if spk1 ==None or spk2==None:
                    import pdb
                    pdb.set_trace()
                answer = Answer(spk1 == spk2, "same", time_1, time_2)
                #answer = {
                #    "answer": {"value": spk1 == spk2},
                #    "response_type": "same",
                #    "time_1": time_1,
                #    "time_2": time_2,
                #}
            elif message.system_request.request_type == "boundary":
                st, en = self.find_segment_for_time(message.system_request.time_1)
                if st != None:
                    self.cost += en - st + self.request_collar_cost
                    print(
                        "USER: Check for boundary on %f (%f - %f)"
                        % (message.system_request.time_1, st, en)
                    )
                    answer = Answer(True, "boundary", st, en )
                    #answer = {
                    #    "answer": {"value": True},
                    #    "response_type": "boundary",
                    #    "time_1": st,
                    #    "time_2": en,
                    #}
                else:
                    self.cost += self.request_collar_cost
                    print(
                        "USER: Check for boundary on %f (not speech)"
                        % (message.system_request.time_1)
                    )
                    answer = Answer(False, "boundary", 0.0, 0.0 )
                    #answer = {
                    #    "answer": {"value": False},
                    #    "response_type": "boundary",
                    #    "time_1": 0.0,
                    #    "time_2": 0.0,
                    #}

        elif message.file_info.supervision == "interactive":
            #print("Interactive model ON")
            answer = find_best_information(self.reference,
                                           message.hypothesis,
                                           0.250,
                                           correction_type=self.correction_type)

        else:
            answer = Answer(True, "stop", 0.0, 0.0 )
            #answer = {
            #    "response_type": 'stop',
            #    "time_1": np.float32(0),
            #    "time_2": np.float32(0),
            #    "answer": {"value": False},
            #}
        self.cost += self.compute_answer_cost(answer)
        #return answer['response_type'] != "stop", answer
        return answer.response_type != "stop", answer

    def find_segment_for_time(self, time):
        for i, s in enumerate(self.reference.speaker):
            if (
                    time >= self.reference.start_time[i]
                    and time < self.reference.end_time[i]
            ):
                return self.reference.start_time[i], self.reference.end_time[i]
        return None, None

    def find_speaker_for_time(self, time):
        for i, s in enumerate(self.reference.speaker):
            #            print(i, s, time, )
            if (
                    time >= self.reference.start_time[i]
                    and time < self.reference.end_time[i]
            ):
                return s
        return None


class Request:

    def __init__(self, request_type, time_1, time_2):
        self.request_type = request_type
        self.time_1 = time_1
        self.time_2 = time_2


class MessageToUser:

    def __init__(self, file_info, hypothesis, system_request):
        self.file_info = file_info
        self.hypothesis = hypothesis
        self.system_request = system_request

class Answer:

    def __init__(self, answer, response_type, time_1, time_2):
        self.answer = answer
        self.response_type = response_type
        self.time_1 = time_1
        self.time_2 = time_2

    def __repr__(self):
        return f"answer = {self.answer}\nresponse_type = {self.response_type}\ntime_1 = {self.time_1}\ntime_2 =  {self.time_2}"

class FileInfo:

    def __init__(self, file_id, supervision, time_stamp):
        self.file_id = file_id
        self.supervision = supervision
        self.time_stamp = time_stamp


class Reference:

    def __init__(self, speaker, start_time, end_time):
        self.speaker = speaker
        self.start_time = start_time
        self.end_time = end_time



class UEM:

    def __init__(self, start_time, end_time):
        self.start_time = start_time
        self.end_time = end_time

